package Forest::Dependency::Writer;

require 5.005_62;
use strict;
use warnings;
use Forest;
use XML::Generator;

our $VERSION = '0.01';
our $forest;
our $dependency;
our %done;

our %color = ( 'subst' => 'blue',
	       'adj' => 'red',
	       'anchor' => 'yellow',
	       'lexical' => 'magenta'
	      );

our %trees;

sub cluster {
    my $l = shift;
    my $r = shift;
    return "cluster$l"."_"."$r";
}

sub node {
    my $l = shift;
    my $r = shift;
    my $cat = shift;
    my $tree = shift;
    return cluster($l,$r)."_$cat"."_".norm_tree(treename($tree));
}

sub treename {
  my $treename = shift;
  return (split(/\s+/,$treename))[0];
}

sub Forest::dependency_normalize {
    my $self=shift;
    my $nb = $self->{'nb'};
    $forest = $self;
    $dependency = {};
    %done=();
    %trees=();

    # First pass
    foreach my $head (keys %{$self->{'op'}}) {
	next if (defined $done{$head});
	$self->{'op'}{$head}->dependency();
    }

    # Second pass: emiting
    foreach my $pos (sort {$a <=> $b} keys %{$dependency}) {
	foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$pos}}) {
	    my $lex = $dependency->{$pos}{$posb}{'lex'};
	    my @trees = keys %{$dependency->{$pos}{$posb}{'tree'}};
	    foreach my $tree (sort @trees) {
		my $cat = $trees{$tree}{'cat'};
		my $node = $trees{$tree}{'node'};
		my $edges = $dependency->{$pos}{$posb}{'tree'}{$tree};
		my $treename = treename($tree);
		foreach my $label (sort keys %{$edges}) {
		    foreach my $target (@{$edges->{$label}}) {
			my $tpos = $target->{'pos'}[0];
			my $tposb = $target->{'pos'}[1];
			my $ttree = $target->{'tree'};
			my $ttreename = treename($ttree);
			my $tlex = $dependency->{$tpos}{$tposb}{'lex'};
			my $tcat = $trees{$ttree}{'cat'};
			my $tnode = $trees{$ttree}{'node'};
			print "$target->{'type'}\t{$pos $posb $lex $treename $node $cat}\t{$tpos $tposb $tlex $ttreename $tnode $tcat}\t$label\n";
		    }
		}
	    }
	}
    }
}

sub tree_emit {
  my ($pos,$posb,$cat,$tree,$grammar) = @_;
  my $treename = treename($tree);
  my $node = node($pos,$posb,$cat,$tree);
  my $url = (defined $grammar && defined $treename) ? ", URL=\"$grammar?id=$treename\"" : "";
  $treename ||= "";
  return <<EOF;
$node [label="$cat:$treename" $url]
EOF
}

sub Forest::collect_dependency {
  my $self = shift;
  $dependency = {};
  $forest = $self;
  %done=();
  %trees=();

  foreach my $head (keys %{$self->{'op'}}) {
    next if (defined $done{$head});
    $self->{'op'}{$head}->dependency();
  }

}

sub Forest::xmldep {
    my $self=shift;

    # First pass: collecting dependencies
    $self->collect_dependency;

    my %nodes = ();
    my $nid= "n001";
    my $eid = "e001";
    my $GEN= new XML::Generator('encoding' => 'ISO-8859-1', 
				':pretty' => 1 );
    my @clusters = ();
    my @nodes = ();
    my @edges = ();

    my $xmlnode = sub {
      my ($posa,$posb,$cat,$tree) = @_;
      unless (exists $nodes{$posa}{$posb}{$cat}{$tree}) {
	my $cluster = "c_".$posa."_".$posb;
	push(@nodes,
	     $GEN->node({ id => $nid,
			  cluster => $cluster,
			  tree => $tree || 'void',
			  cat => $cat
			}));
	$nodes{$posa}{$posb}{$cat}{$tree}{id} = $nid;
	++$nid;
      }
      return $nodes{$posa}{$posb}{$cat}{$tree}{id}
    };

    # Second pass: build xml
    foreach my $pos (sort {$a <=> $b} keys %{$dependency}) {
      foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$pos}}) {
	push(@clusters, $GEN->cluster({ id=> "c_".$pos."_".$posb,
					left => $pos,
					right => $posb,
					lex =>  $dependency->{$pos}{$posb}{'lex'} || ""
				      }));
	foreach my $cat (keys %{$dependency->{$pos}{$posb}{'tree'}}) {
	  foreach my $tree (keys %{$dependency->{$pos}{$posb}{'tree'}{$cat}}) {
	    my $edges = $dependency->{$pos}{$posb}{'tree'}{$cat}{$tree};
	    my $ida = $xmlnode->($pos,$posb,$cat,$tree);
	    foreach my $label (keys %{$edges}) {
	      foreach my $target (@{$edges->{$label}}) {
		my $idb = $xmlnode->( $target->{'pos'}[0],
				      $target->{'pos'}[1],
				      $target->{cat},
				      $target->{'tree'});
		push(@edges,$GEN->edge({target =>$ida,
					source =>$idb,
					id =>$eid++,
					type => $target->{type},
					label => $label
				       }));
	      }
	    }
	  }
	}
      }
    }

    # Third pass: emit
    print $GEN->xmldecl;
    print $GEN->dependencies({ grammar => $self->{grammar},
			       sentence => $self->{sentence}
			     },
			     @clusters,
			     @nodes,
			     @edges), "\n";

}


sub lp_quote {
  my $string = shift || "";
  return $string if ($string =~ /^[a-z_]\w*$/o);
  return $string if ($string =~ /^\d+$/);
  #    $string =~ s/\'/\'\'/og;
  $string =~ s/\'(?!\')/\'\'/og;
  return "\'$string\'";
}

sub Forest::lpdep {
    my $self=shift;

    # First pass: collecting dependencies
    $self->collect_dependency;

    my %nodes = ();
    my $nid= "n001";
    my $eid = "e001";

    my $lpnode = sub {
      my ($posa,$posb,$cat,$tree) = @_;
      unless (exists $nodes{$posa}{$posb}{$cat}{$tree}) {
	my $cluster = "c_".$posa."_".$posb;
	my $qtree = lp_quote(treename($tree || 'void'));
	my $qcat = lp_quote($cat);
	print "node{ id => $nid, cluster => $cluster, tree => $qtree, cat => $qcat }.\n"; 
	$nodes{$posa}{$posb}{$cat}{$tree}{id} = $nid;
	++$nid;
      }
      return $nodes{$posa}{$posb}{$cat}{$tree}{id}
    };

    print "%% sentence: $self->{sentence}\n" if (exists $self->{sentence});
    print "%% parser: $self->{parser}\n" if (exists $self->{parser});

    # Second pass: emit
    foreach my $pos (sort {$a <=> $b} keys %{$dependency}) {
      foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$pos}}) {
	my $qlex = lp_quote($dependency->{$pos}{$posb}{'lex'});
	my $cluster = "c_".$pos."_".$posb;
	print "cluster{ id => $cluster, left => $pos, right => $posb, lex => $qlex }.\n";
	foreach my $cat (keys %{$dependency->{$pos}{$posb}{'tree'}}) {
	  foreach my $tree (keys %{$dependency->{$pos}{$posb}{'tree'}{$cat}}) {
	    my $edges = $dependency->{$pos}{$posb}{'tree'}{$cat}{$tree};
	    my $ida = $lpnode->($pos,$posb,$cat,$tree);
	    foreach my $label (keys %{$edges}) {
	      foreach my $target (@{$edges->{$label}}) {
		my $idb = $lpnode->( $target->{'pos'}[0],
				      $target->{'pos'}[1],
				      $target->{cat},
				      $target->{'tree'});
		my $qlabel = lp_quote($label);
		my $ide = $eid++;
		print "edge{ target => $ida, source => $idb, id => $ide, type => $target->{type}, label => $qlabel }.\n";
	      }
	    }
	  }
	}
      }
    }

}


sub Forest::stats {
    my $self=shift;
    
    # First pass: collecting dependencies
    $self->collect_dependency;

    # Second pass: compute stats
    my $sclusters = 0;
    my $snodes = 0;
    my $sedges = 0;

    foreach my $pos (sort {$a <=> $b} keys %{$dependency}) {
      foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$pos}}) {
	$sclusters++;
	foreach my $cat (keys %{$dependency->{$pos}{$posb}{'tree'}}) {
	  foreach my $tree (keys %{$dependency->{$pos}{$posb}{'tree'}{$cat}}) {
	    $snodes++;
	    my $edges = $dependency->{$pos}{$posb}{'tree'}{$cat}{$tree};
	    foreach my $label (keys %{$edges}) {
	      foreach my $target (@{$edges->{$label}}) {
		$sedges++;
	      }
	    }
	  }
	}
      }
    }

    printf "Dependency stats: %2.1f ambiguity %3d clusters %3d nodes %3d edges\n",
      ($sedges - $sclusters) / ($sclusters || 1),
	$sclusters,
	  $snodes,
	    $sedges;

}

sub Forest::dependency {
    my $self=shift;
    my $nb = $self->{'nb'};
    my $grammar = $self->{'grammar'};
    $forest = $self;

    # First pass: collecting dependencies
    $self->collect_dependency;

    # Second pass: emiting
    if (defined $grammar) {
      print "/* Grammar=$grammar */\n";
    }

    print <<HEADER;
digraph G {
    rankdir = LR;
    ranksep = .5;
    ratio = auto;
    clusterrank= local;
HEADER

    foreach my $pos (sort {$a <=> $b} keys %{$dependency}) {
      foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$pos}}) {
	my $lex = $dependency->{$pos}{$posb}{'lex'};
	my $cluster = cluster($pos,$posb);
	my @trees = ();

	foreach my $cat (keys %{$dependency->{$pos}{$posb}{'tree'}}) {

	  ## Emit edges
	  foreach my $tree (keys %{$dependency->{$pos}{$posb}{'tree'}{$cat}}) {
	    push(@trees,tree_emit($pos,$posb,$cat,$tree,$grammar));
	    #	    print "/* $cluster $tree */\n";
	    my $edges = $dependency->{$pos}{$posb}{'tree'}{$cat}{$tree};
	    my $start = node($pos,$posb,$cat,$tree);
	    foreach my $label (keys %{$edges}) {
	      #		print "/* label = $label */\n";
	      foreach my $target (@{$edges->{$label}}) {
		#		    print "/* end = $target->{'pos'} $target->{'tree'} */\n";
		my $end = node( $target->{'pos'}[0],
				$target->{'pos'}[1],
				$target->{cat},
				$target->{'tree'} );
		my $color = $color{$target->{'type'}};
		my $weight = abs($target->{'pos'}[0]-$pos);
		$weight = 1 unless ($weight);
		if ($target->{'pos'}[0] > $pos) {
		  print <<EDGE;
     $start -> $end [ label = "$label", color = "$color", minlen = $weight, dir = "back" ];
EDGE
		} else {
		  print <<EDGE;
     $end -> $start [ label = "$label", color = "$color",  minlen = $weight ];
EDGE
		}
	      }
	    }
	  }
	}

	## Emit nodes
	my $trees = join(";\t\n",@trees);
	my $qlex = $lex;
	$qlex = "\\$qlex" if ($qlex eq '"');
	print <<ANCHOR;
subgraph  $cluster {
	    fontcolor = "magenta";
	    label = "$qlex";
	    {rank = same; $trees; }
	}
ANCHOR

      }
    }
    print "}\n\n";
  }

sub Forest::Op::dependency {
    # return a list of anchors
    my $self=shift;
    my $tree= shift;
    my $head = $self->{'id'};
    return $done{$head} if ($done{$head});
    my $type = $self->{type};
    if ($type eq 'anchor' || $type eq 'lexical') {
      return unless (defined $tree);
      my $cat = $self->{cat};
      my ($left,$right) = @{$self->{'span'}};
      my $entry =  $dependency->{$left}{$right} ||= {};
      $entry->{lex} ||= $self->{'lex'};
##      print "\nWARNING $type TREE ANCHOR $self->{tree} vs DERIV $tree\n";
##      $entry->{tree}{$cat}{$self->{tree}} ||= {};
      my $xtree = ($type eq 'lexical') ? "" : $tree;
      $entry->{tree}{$cat}{$xtree} ||= {};
      return $done{$head} = [ $self ];
    } else {
      my @anchors = ();		# the anchors of the various derivations for op
      foreach (@{ $self->{'derivations'}}) {
	push(@anchors,$_->dependency($head));
      }
      return $done{$head} = [ @anchors ];
    }
}

sub Forest::Deriv::dependency {
    # return an anchor
    my $self=shift;
    my $op=shift;
    my @anchors;
    my @edges = ();
    my $tree = $self->{tree};
    $trees{$tree} = { 'cat' => $forest->{'op'}{$op}->{'cat'} };
    foreach my $label (keys %{$self->{'nodes'}}){
	foreach my $node (@{$self->{'nodes'}{$label}}){
	    next unless (defined $forest->{'op'}{$node});
	    my $target = $forest->{'op'}{$node};
	    if ($target->{'type'} eq 'anchor' && $target->{'span'}[0] < $target->{'span'}[1] ){
	      push(@anchors,$target);
	      $trees{$tree}{node} = $label;
	      next ;
	    }
	    if ($label =~ /^\d+/) {
		my $cat =  $target->{'cat'};
		$label = "$cat\#$label";
	    }
	    push(@edges, { 'type' => $target->{'type'},
			   'label' => $label,
			   'anchors' => $target->dependency($tree)
			   } 
		 );
	}
    }
    ## WARNING: we may have cases of derivation with no anchor and no indication of tree
    ## what should we do ? maybe use the anchor of some node in deriv as anchor
    if (!@anchors) {
      ## We create a pseudo anchor
      my $xop = $forest->{'op'}{$op};
      my $dummy = Forest::Op->new( id => "$tree @{$xop->{span}}",
				   type => 'anchor',
				   lex => "",
				   derivations => [],
				   tree => $tree,
##				   span => [$xop->{span}[0],$xop->{span}[1]],
				   span => [$xop->{span}[0],$xop->{span}[0]],
				   cat => $xop->{cat} );
      $dummy->dependency($tree);
      push(@anchors,$dummy);
    }
    my $anchor = $anchors[0];
    $anchor->dependency($tree);
    foreach my $edge (@edges) {
	foreach my $start (@{$edge->{'anchors'}}) {
	  ## next unless $start;
	    $start->dependency();
	    my $sdep = $dependency->{$start->{'span'}[0]}{$start->{'span'}[1]}{'tree'};
	    $sdep = $sdep->{$start->{'cat'}}{$start->{'tree'} || ""};
#	    print "/* add edge $edge->{'type'} from $start->{'tree'} $start->{'span'}[0] to  $anchor->{'tree'} $anchor->{'span'}[0] with label $edge->{'label'} */\n";
	    $sdep->{$edge->{'label'}} ||= [];
	    my $info = { type => $edge->{'type'},
			 pos => $anchor->{'span'},
			 tree => $tree,
			 cat => $anchor->{cat}
		       };
	    next if (grep(&is_edge($_,$info),@{$sdep->{$edge->{'label'}}}));
	    push(@{$sdep->{$edge->{'label'}}},$info);
	}
    }
    return $anchor;
}

sub is_edge {
    my $e1 = shift;
    my $e2 = shift;
    return ($e1->{'type'} eq $e2->{'type'} 
	    && $e1->{'pos'}[0] == $e2->{'pos'}[0] 
	    && $e1->{'pos'}[1] == $e2->{'pos'}[1] 
	    && $e1->{'cat'} eq $e2->{'cat'}
	    && $e1->{'tree'} eq $e2->{'tree'});
}

sub norm_tree {
    my $tree = shift || "";
#    print "/* norm tree $tree */\n";
    $tree =~ tr/-/_/;
    $tree =~ s/\s+/_/og;
    return $tree;
}


1;

__END__

=head1 NAME

Forest - Perl Objects to represent TAG Derivation Forest

=head1 SYNOPSIS

  use Forest;

=head1 DESCRIPTION

=head2 EXPORT

None by default.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=head1 SEE ALSO

perl(1).

=cut
